# -*- coding: utf-8 -*-

"""Utility functions for a2r

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2019, G.J.J. van den Burg

"""

import os
import string
import subprocess
import time

import regex
import requests
import unidecode

from pikepdf import Pdf
from pikepdf import PdfError

from .exceptions import FileTypeError
from .exceptions import NoPDFToolError
from .exceptions import RemarkableError
from .log import Logger

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 "
    "Safari/537.36"
}


logger = Logger()


def clean_string(s):
    """Clean a string by replacing accented characters with equivalents and
    keeping only the allowed characters (ascii letters, digits, underscore,
    space, dash, and period)"""
    normalized = unidecode.unidecode(s)
    allowed = string.ascii_letters + string.digits + "_ .-"
    cleaned = "".join(c if c in allowed else "_" for c in normalized)
    while "__" in cleaned:
        cleaned = cleaned.replace("__", "_")
    cleaned = cleaned.strip("_")
    return cleaned


def assert_file_is_pdf(filename):
    """Assert that a given file is a PDF file.

    This is done by trying to open it using pikepdf.
    """
    try:
        pdf = Pdf.open(filename)
        pdf.close()
        del pdf
        return True
    except PdfError:
        raise FileTypeError(filename, "pdf")


def download_url(url, filename, cookiejar=None):
    """Download the content of an url and save it to a filename"""
    logger.info("Downloading file at url: %s" % url)
    content = get_page_with_retry(url, cookiejar=cookiejar)
    with open(filename, "wb") as fid:
        fid.write(content)


def get_page_with_retry(url, tries=5, cookiejar=None, return_text=False):
    count = 0
    jar = {} if cookiejar is None else cookiejar
    while count < tries:
        count += 1
        error = False
        try:
            res = requests.get(url, headers=HEADERS, cookies=jar)
        except requests.exceptions.ConnectionError:
            error = True
        if error or not res.ok:
            logger.warning(
                "(%i/%i) Error getting url %s. Retrying in 5 seconds."
                % (count, tries, url)
            )
            time.sleep(5)
            continue
        logger.info("Downloaded url: %s" % url)
        if return_text:
            return res.text
        return res.content


def get_content_type_with_retry(url, tries=5, cookiejar=None):
    if cookiejar is None:
        jar = requests.cookies.RequestsCookieJar()
    else:
        jar = cookiejar

    msg = "(%i/%i) Error getting content type for %s. Retrying in 5 seconds."

    # In rare cases, a HEAD request fails but a GET request does work. So here
    # we try both
    ops = [requests.head, requests.get]
    kwargs = dict(headers=HEADERS, cookies=jar, allow_redirects=True)
    for op in ops:
        count = 0
        while count < tries:
            count += 1
            error = False
            try:
                res = op(url, **kwargs)
            except requests.exceptions.ConnectionError:
                error = True
            if error or not res.ok:
                logger.warning(msg % (count, tries, url))
                time.sleep(5)
                continue
            return res.headers.get("Content-Type", None)
    return None


def follow_redirects(url):
    """Follow redirects from the URL (at most 100)"""
    it = 0
    jar = requests.cookies.RequestsCookieJar()
    while it < 100:
        req = requests.head(
            url, headers=HEADERS, allow_redirects=False, cookies=jar
        )
        if req.status_code == 200:
            break
        if not "Location" in req.headers:
            break
        url = req.headers["Location"]
        jar.update(req.cookies)
        it += 1
    if it == 100:
        logger.warning("Max redirects reached. There may be a problem.")
    jar = jar or req.cookies
    return url, jar


def upload_to_remarkable(filepath, remarkable_dir="/", rmapi_path="rmapi"):
    logger.info("Starting upload to reMarkable")

    # Create the reMarkable dir if it doesn't exist
    remarkable_dir = remarkable_dir.rstrip("/")
    if remarkable_dir:
        parts = remarkable_dir.split("/")
        rmdir = ""
        while parts:
            rmdir += "/" + parts.pop(0)
            status = subprocess.call(
                [rmapi_path, "mkdir", rmdir],
                stdout=subprocess.DEVNULL,
            )
            if not status == 0:
                raise RemarkableError(
                    "Creating directory %s on reMarkable failed"
                    % remarkable_dir
                )

    # Upload the file
    status = subprocess.call(
        [rmapi_path, "put", filepath, remarkable_dir + "/"],
        stdout=subprocess.DEVNULL,
    )
    if not status == 0:
        raise RemarkableError(
            "Uploading file %s to reMarkable failed" % filepath
        )
    logger.info("Upload successful.")


def is_url(string):
    # pattern adapted from CleverCSV
    pattern = "((https?|ftp):\/\/(?!\-))?(((([\p{L}\p{N}]*[\-\_]?[\p{L}\p{N}]+)+\.)+([a-z]{2,}|local)(\.[a-z]{2,3})?)|localhost|(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}(\:\d{1,5})?))(\/[\p{L}\p{N}_\/()~?=&%\-\#\.:+]*)?(\.[a-z]+)?"
    string = string.strip(" ")
    match = regex.fullmatch(pattern, string)
    return match is not None


def check_pdftool(pdftk_path, qpdf_path):
    """Check whether we have pdftk or qpdf available"""
    # set defaults in case either is set to None or something
    pdftk_path = pdftk_path or "false"
    qpdf_path = qpdf_path or "false"

    try:
        status = subprocess.call(
            [pdftk_path], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
    except FileNotFoundError:
        status = 1
    if status == 0:
        return "pdftk"
    try:
        status = subprocess.call(
            [qpdf_path, "--help"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        status = 1
    if status == 0:
        return "qpdf"
    raise NoPDFToolError


class chdir:
    """Change directory in context and return to original on exit or failure"""

    def __init__(self, target: str):
        self._target = target
        self._original_dir = None

    def __enter__(self):
        self._original_dir = os.getcwd()
        os.chdir(self._target)

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self._original_dir)



# -*- coding: utf-8 -*-

"""Provider for ScienceDirect

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2020, G.J.J. van den Burg

"""

import json
import re
import urllib

import bs4

from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Util.Padding import pad

from ..exceptions import URLResolutionError
from ..log import Logger
from ..utils import get_page_with_retry
from ._base import Provider
from ._info import Informer

logger = Logger()


class ScienceDirectInformer(Informer):

    meta_date_key = "citation_publication_date"

    def get_authors(self, soup):
        surname_tags = soup.find_all("span", attrs={"class": "text surname"})
        if not surname_tags:
            logger.warning(
                "Couldn't determine author information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        authors = [x.text for x in surname_tags]
        return authors


class ScienceDirect(Provider):

    re_abs = (
        "https?:\/\/www.sciencedirect.com/science/article/pii/[A-Za-z0-9]+"
    )
    re_pdf = "https://pdf.sciencedirectassets.com/\d+/([0-9a-zA-Z\-\.]+)/(?P<data>[0-9a-zA-Z\-\.]+)/main.pdf\?.*"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = ScienceDirectInformer()

    def get_abs_pdf_urls(self, url):
        m1 = re.match(self.re_abs, url)
        m2 = re.match(self.re_pdf, url)
        if m1:
            abs_url = url
            pdf_url = self._get_pdf_url(abs_url)
        elif m2:
            pdf_url = url
            data = m2.group("data")
            paper_id = data.split("-")[-1]
            abs_url = (
                f"https://www.sciencedirect.com/science/article/pii/{paper_id}"
            )
        else:
            raise URLResolutionError("ScienceDirect", url)
        return abs_url, pdf_url

    def _get_pdf_url(self, url):
        page = get_page_with_retry(url)
        soup = bs4.BeautifulSoup(page, "html.parser")

        # For open access (and maybe behind institution?) the full text pdf url
        # is retrieved by mimicking the authentication dance in Python.

        # Extract article metadata from the page
        scripts = soup.find_all("script", attrs={"data-iso-key": "_0"})
        if not scripts:
            raise URLResolutionError("ScienceDirect", url)
        json_data = scripts[0].string
        data = json.loads(json_data)
        if not "article" in data:
            raise URLResolutionError("ScienceDirect", url)

        data = data["article"]
        if not "pdfDownload" in data:
            raise URLResolutionError("ScienceDirect", url)

        data = data["pdfDownload"]

        if not "urlMetadata" in data:
            raise URLResolutionError("ScienceDirect", url)
        meta = data["urlMetadata"]

        # construct a temporary url to an intermediate page
        link = "{path}/{pii}/{pdfExtension}?md5{queryParams[md5]}&pid={queryParams[pid]}".format(
            **meta
        )
        tmp_url = urllib.parse.urljoin("https://sciencedirect.com/", link)

        # Open the temp url, this lands us on a page that requires Javascript
        # to do an authentication dance
        page = get_page_with_retry(tmp_url)
        soup = bs4.BeautifulSoup(page, "html.parser")
        script = soup.find(
            lambda tag: tag.name == "script" and "subtle.encrypt" in tag.text
        )
        if script is None:
            raise URLResolutionError(
                "ScienceDirect", url, reason="Can't find script element"
            )
        script_text = script.decode()

        # Extract the embedded information from the script
        phrase_1 = 'subtle.digest("SHA-256",'
        offset_1 = len(phrase_1) + 3
        try:
            rem = script_text[script_text.index(phrase_1) + offset_1 :]
            token = rem[: rem.index('"')]
        except ValueError as err:
            raise URLResolutionError(
                "ScienceDirect", url, reason="Can't find encryption token"
            ) from err

        phrase_2 = "subtle.encrypt("
        offset_2 = len(phrase_2) + 7
        try:
            rem = script_text[script_text.index(phrase_2) + offset_2 :]
            data = rem[: rem.index('"')]
        except ValueError as err:
            raise URLResolutionError(
                "ScienceDirect", url, reason="Can't find encryption data"
            ) from err

        phrase_3 = 'window.location="'
        try:
            rem = script_text[script_text.index(phrase_3) + len(phrase_3) :]
            location = rem[: rem.index("}") - len('",r()')]
        except ValueError as err:
            raise URLResolutionError("ScienceDirect", url) from err

        location = location.replace('"+e+"', "{e}")
        location = location.replace('"+t+"', "{t}")

        # Perform the authentication dance in Python
        e, t = self.sd_run(token, data)
        tmp_url2 = urllib.parse.urljoin(
            "https://www.sciencedirect.com", location.format(e=e, t=t)
        )

        # tmp_url2 gives a page with a ten second wait or a direct url, we need
        # the direct url
        page = get_page_with_retry(tmp_url2)
        soup = bs4.BeautifulSoup(page, "html.parser")
        noscript = soup.find_all("noscript")
        if not noscript:
            raise URLResolutionError("ScienceDirect", url)
        a = noscript[0].find_all("a")
        if not a:
            raise URLResolutionError("ScienceDirect", url)
        pdf_url = a[0].get("href")
        return pdf_url

    def validate(src):
        return re.match(ScienceDirect.re_abs, src) or re.match(
            ScienceDirect.re_pdf, src
        )

    def sd_e(self, e):
        n = []
        for o in range(len(e)):
            n.append(ord(e[o]))
        return n

    def sd_t(self, o):
        t = ""
        n = "0123456789abcdef"
        for r in range(len(o)):
            i = o[r]
            t += n[i >> 4] + n[15 & i]
        return t

    def sd_run(self, token, data):
        # token is the string that is passed to sha-256
        # data is the string that is passed to encrypt
        a = Random.new().read(16)
        d = self.sd_t(a)

        h = SHA256.new()
        h.update(bytearray(self.sd_e(token)))
        key = list(h.digest())

        aes = AES.new(bytearray(key), AES.MODE_CBC, iv=a)
        msg = bytearray(self.sd_e(data))
        ct_bytes = aes.encrypt(pad(msg, AES.block_size))
        ct_uint8 = list(ct_bytes)
        return d, self.sd_t(ct_uint8)



# -*- coding: utf-8 -*-

"""Provider for Electronic Colloquium on Computational Complexity

Author: G.J.J. van den Burg
License: See LICENSE file
Copyright: 2021, G.J.J. van den Burg

"""

import re

import bs4

from ..exceptions import URLResolutionError
from ..log import Logger
from ._base import Provider
from ._info import Informer

logger = Logger()


class ECCCInformer(Informer):
    def _get_paper_div(self, soup):
        h3 = soup.find(lambda t: t.name == "h3" and t.get_text() == "Paper:")
        div = h3.find_next_sibling("div")
        return bs4.BeautifulSoup(div.prettify(), "html.parser")

    def get_title(self, soup):
        divsoup = self._get_paper_div(soup)
        h4 = divsoup.find("h4")
        if not h4:
            logger.warning(
                "Couldn't determine title information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        return h4.get_text().strip()

    def get_authors(self, soup):
        divsoup = self._get_paper_div(soup)
        aa = divsoup.find_all(
            lambda t: t.name == "a" and t.get("href").startswith("/author/")
        )
        if not aa:
            logger.warning(
                "Couldn't determine author information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        authors = [a.get_text() for a in aa]
        return self._format_authors(authors, sep=" ", idx=-1)

    def get_year(self, soup):
        divsoup = self._get_paper_div(soup)
        line = next(
            (l for l in divsoup.text.split("\n") if "Publication: " in l), None
        )
        if line is None:
            logger.warning(
                "Couldn't determine year information, maybe provide the desired filename using '--filename'?"
            )
            return ""
        year = line.strip().split(" ")[3]  # bit lazy
        return year


class ECCC(Provider):

    re_abs = "https?://eccc.weizmann.ac.il/report/\d{4}/\d+/?$"
    re_pdf = "https?://eccc.weizmann.ac.il/report/\d{4}/\d+/download/?$"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.informer = ECCCInformer()

    def get_abs_pdf_urls(self, url):
        if re.match(self.re_abs, url):
            abs_url = url
            pdf_url = url.rstrip("/") + "/download"
        elif re.match(self.re_pdf, url):
            abs_url = url.rstrip("/")[: -len("/download")]
            pdf_url = url
        else:
            raise URLResolutionError("ECCC", url)
        return abs_url, pdf_url

    def validate(src):
        return re.match(ECCC.re_abs, src) or re.match(ECCC.re_pdf, src)
